import os
from typing import List


# https://www.sethserver.com/python/recursively-list-files.html


def get_all_files_recursively_from_directory(dirname: str) -> List[str]:
    """
    Get all files recursively from a directory.

    :param dirname: The directory name.
    :type dirname: str
    :return: The list of founded files.
    :rtype: List[str]
    """
    files = []
    for (dirpath, dirnames, filenames) in os.walk(dirname):
        files.extend(
            map(
                lambda n: os.path.join(*n),
                zip([dirpath] * len(filenames), filenames),
            )
        )
    return files



#!/usr/bin/env python3

"""
A chronometer to schedule task.

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

import time

import arrow


class OneMinuteChronometer(object):

    def __init__(self):
        self.current, self.next = arrow.now(), arrow.now()
        self.set_next()

    def set_next(self):
        self.next = arrow.now().replace(second=0, microsecond=0).shift(minutes=+1)

    def check(self):
        self.update_current()
        if self.current > self.next:
            self.set_next()
            return True
            # return self.current.replace(second=0, microsecond=0).format('YYYYMMDDhhmm')
        return False


    def update_current(self):
        self.current = arrow.now()

if __name__ == '__main__':
    chrono = OneMinuteChronometer()
    while True:
        time.sleep(5)
        c = chrono.check()
        if c is not None:
            print(c)


#!/usr/bin/env python3

"""

# Module: loadguard.core.web.url

This file is a part of LoadGuard Runner.

(c) 2021, Deepnox SAS.

This module provides utilities to manage URLs.

"""

from typing import Dict, Optional, Union, Any
from urllib.parse import ParseResult, urlencode

from deepnox import loggers
from deepnox.models import ExtendedBaseModel
from deepnox.third import pydantic
from deepnox.network import Scheme
from deepnox.utils.strings_utils import remove_slash_at_start_and_at_end

LOGGER = loggers.factory(__name__)
""" The main LOGGER. """


class Url(ExtendedBaseModel):
    """
    Url entity.

    Encapsulation of Python :class:`urllib.parse.ParseResult`.
    """

    scheme: Optional[Scheme] = None
    """ The scheme. """

    hostname: Optional[str] = None
    """ The hostname. """

    port: Optional[int] = None
    """ The port. """

    path: Optional[str] = None
    """ The path. """

    # __attrs__ = ['scheme', 'netloc', 'path', 'params', 'query', 'hostname', 'port']
    # __attrs__ = ["scheme", "hostname", "port", "path"]

    def to_python(self) -> ParseResult:
        """
        Convert instance to native `class`:urllib.parser.ParseResult: object.

        :return:
        :rtype: :class:`urllib.parser.ParseResult`
        """
        return ParseResult(self.scheme, self.netloc, self.path, self.params, self._query, None)

    @pydantic.validator('scheme', pre=True, always=True)
    def scheme_autoconvert_or_default(cls, v):
        if isinstance(v, Scheme):
            return v
        if isinstance(v, str):
            return Scheme(v)

    # @property
    # def netloc(self):
    #     if self.port:
    #         return f"{self.host}:{self.port}"
    #     return self.host

    def __str__(self):
        url = f"{str(self.scheme)}://" + "/".join(
            [remove_slash_at_start_and_at_end(self.hostname), remove_slash_at_start_and_at_end(self.path)])
        # url = self._add_query_string_params_to_url(url)
        return url

    def __repr__(self):
        return f'<Url ({str(self)})>'

    # def _add_query_string_params_to_url(self, url: str):
    #     if self.params is not None and len(self.params.keys()) > 0:
    #         return f"{str(self.scheme)}://{self.hostname}/{self.path}?{urlencode(self.params)}"
    #     return url

    def dict(
            self,
            **kwargs
    ) -> Dict[str, Any]:
        kwargs.update({"exclude_none": True})
        return super().dict(**kwargs)

    class Config:
        json_encoders = {
            Scheme: lambda s: str(s),
        }
        dict_encoders = {
            Scheme: lambda s: str(s),
        }
        extra = pydantic.Extra.forbid



#!/usr/bin/env python3

"""
Base of (deepnox.aiobox) module.

Module: deepnox.aiobox.base

This file is a part of (deepnox.aiobox) project.

(c) 2021, Deepnox SAS.
"""
import asyncio
from types import FunctionType


class Task(object):
    """
    A task.
    """

    def __init__(self,
                 loop: asyncio.AbstractEventLoop = None,
                 fn: FunctionType = None,
                 ):
        if not isinstance(fn, FunctionType):
            raise TypeError(f"Routine cannot be typed: {type(fn)}")
        self.loop = loop or asyncio.get_event_loop()
        self.fn = fn



#!/usr/bin/env python3

"""
Module: deepnox.clients.http_client

This file is a part of (deepnox.clients) project.

(c) 2021, Deepnox SAS.
"""

import asyncio
import logging
import time
import urllib
from types import FunctionType

from deepnox import loggers
from deepnox.network.http import HttpRequest, HttpResponse, HttpHit, HttpMethod, HttpRequestPayload
from deepnox.third import aiohttp

LOGGER = loggers.factory(__name__)
""" The module LOGGER. """


class HttpClient(object):
    """
    The HTTP client.

    """

    LOG = LOGGER.getChild("HttpClient")
    """ The LOGGER. """

    def __init__(self,
                 loop: asyncio.AbstractEventLoop = None,
                 auditor_logger=None,
                 timeout: int = 30,
                 verify_ssl: bool = False
                 ):
        self.loop = loop or asyncio.get_event_loop()
        self.AUDITOR = auditor_logger or loggers.auditor(f'auditor')

    @property
    def get_session(self):
        def _wrap():
            cookie_jar = aiohttp.CookieJar()
            timeout = aiohttp.ClientTimeout(total=None, sock_connect=30, sock_read=30)
            connector = aiohttp.TCPConnector(ssl=False)
            return aiohttp.ClientSession(cookie_jar=cookie_jar, timeout=timeout, connector=connector)
        return _wrap

    async def _parse_response(self, req: HttpRequest, resp: HttpResponse):
        """
        Wait the HTTP response.

        :param req: The HTTP request.
        :param resp: The HTTP response.
        :return:
        """
        logging.info(f"resp {resp}")

        text = await resp.text()
        status_code = resp.status
        end_at = time.time()
        response = HttpResponse(status_code=status_code,
                                headers=resp.headers,
                                text=text,
                                end_at=end_at,
                                elapsed_time=end_at - req.start_at)
        return response

    def _build_request_args(self, req: HttpRequest):
        data = {"url": str(req.url)}
        if isinstance(req.headers, dict):
            data.update({"headers": req.headers})
        if isinstance(req.payload, HttpRequestPayload):
            if isinstance(req.payload.params, dict):
                data.update({"params": req.payload.params})
            if isinstance(req.payload.data, str):
                data.update({"data": req.payload.data})
        return data

    async def request(self, req: HttpRequest):
        """
        Send a HTTP request.
        :param req: The request.
        :type req: :class:`deepnox.network.http.HttpRequest`
        :return:
        """
        LOGGER.debug(f'get(req:{req})', extra={"request": req.dict()})
        method = str(getattr(req, "method"))
        req.start_at = time.time()
        async with self.get_session() as session:
            try:
                self.LOG.debug(f"req.url.to_python() {req.url}")
                session_method_fn: FunctionType = getattr(session, method)
                async with session_method_fn(**self._build_request_args(req)) as resp:
                    res = await self._parse_response(req, resp)
                    self._trace_audit(req, res)
                    return res
            except Exception as e:
                self.LOG.error(f'Response error: (req:{req.url})', exc_info=e)

    async def get(self, req: HttpRequest):
        """
        Send a HTTP request using the verb: GET.
        :param req: The HTTP request.
        :type req: :class:`deepnox.network.http.HttpRequest`
        """
        req.method = HttpMethod.GET
        return self.request(req)

    async def post(self, req: HttpRequest):
        """
        Send a HTTP request using the verb: POST.
        :param req: The HTTP request.
        :type req: :class:`deepnox.network.http.HttpRequest`
        """
        req.method = HttpMethod.POST
        return self.request(req)

    async def put(self, req: HttpRequest):
        """
        Send a HTTP request using the verb: PUT.
        :param req: The HTTP request.
        :type req: :class:`deepnox.network.http.HttpRequest`
        """
        req.method = HttpMethod.PUT
        return self.request(req)

    async def patch(self, req: HttpRequest):
        """
        Send a HTTP request using the verb: PATCH.
        :param req: The HTTP request.
        :type req: :class:`deepnox.network.http.HttpRequest`
        """
        req.method = HttpMethod.POST
        return self.request(req)

    async def options(self, req: HttpRequest):
        """
        Send a HTTP request using the verb: OPTIONS.
        :param req: The HTTP request.
        :type req: :class:`deepnox.network.http.HttpRequest`
        """
        req.method = HttpMethod.OPTIONS
        return self.request(req)

    async def head(self, req: HttpRequest):
        """
        Send a HTTP request using the verb: HEAD.
        :param req: The HTTP request.
        :type req: :class:`deepnox.network.http.HttpRequest`
        """
        req.method = HttpMethod.HEAD
        return self.request(req)

    def _trace_audit(self, req: HttpRequest, res: HttpResponse):
        """
        Add trace.

        :param req: The HTTP request.
        :type req: :class:`deepnox.network.http.HttpRequest`
        :param res: The HTTP response.
        :type res: :class:`deepnox.network.http.HttpResponse`
        :return:
        """
        self.LOG.debug("_trace_audit(req, res)", extra={"req": req.dict(), "res": res.dict()})
        http_hit = HttpHit(start_at=req.start_at, end_at=res.end_at,
                           url=req.url, method=req.method,
                           request=req, response=res)
        if 200 < res.status_code >= 400:
            self.LOG.error(f"Failed: ({str(req.method)} at url:{req.url})", extra=http_hit.dict())
            print(http_hit.json())
        else:
            self.LOG.info(f"Success: ({str(req.method)} at url:{req.url})", extra=http_hit.dict())

    # def __del__(self):
    #     try:
    #         asyncio.create_task(self._close_session())
    #     except RuntimeError:
    #         self.loop.run_until_complete(self._close_session())
    #
    # async def _close_session(self):
    #     if not self._session.closed:
    #         self._session.close()






import re


def get_pgsql_datasource_from_springboot_settings(data: dict = None):
    if data is None:
        raise Exception('Any data')

    datasource = data.get('spring').get('datasource')
    DB_USER = datasource.get('username')
    DB_PASSWORD = datasource.get('password')

    DB_URI = datasource.get('url')  # jdbc:postgresql://pzbdacdh02.postgres.database.azure.com:5432/bacd00
    DB_HOST, DB_PORT, DB_NAME = re.match('jdbc:postgresql://(.*?):(.*?)/(.*)', DB_URI).groups()
    return DB_HOST, DB_PORT, DB_USER, DB_PASSWORD, DB_NAME


#!/usr/bin/env python3

"""
Base classes and functions used by deepnox.tests.repositories.

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

import logging

import pandas as pd
from pandas import DataFrame

from deepnox import loggers

LOGGER: logging.Logger = loggers.factory(__name__)
loggers.setup()


class BaseRepository(object):
    """
    A deepnox.tests.app class for deepnox.tests.repositories.
    """


class Repository(BaseRepository):
    """
    A deepnox.tests.app class for () computable deepnox.tests.repositories.
    """


class ComputableRepository(BaseRepository):
    """
    A deepnox.tests.app class for () computable deepnox.tests.repositories.
    """

    LOG: logging.Logger = LOGGER.getChild("ComputableRepository")
    """ The class LOGGER. """

    def __init__(self, model_cls: object = None, input_data: object = None):
        self._model_cls = model_cls
        self._df: pd.DataFrame = pd.DataFrame()

    def indexes(self):
        """
        Return list containing indexes names.
        :return: The list containing indexes names.
        :rtype: list
        """
        return list(
            filter(
                lambda x: x is not False,
                [
                    v.index and k
                    for k, v in self._model_cls._attributes.items()
                ],
            )
        )  # :see: https://bit.ly/31KwLee

    def primary_keys(self):
        """
        Return list containing primary key(s) names.
        :return: The list primary key(s) names.
        :rtype: list
        """
        return list(
            filter(
                lambda x: x is not False,
                [
                    v.pk is True and k
                    for k, v in self._model_cls._attributes.items()
                ],
            )
        )  # :see: https://bit.ly/31KwLee

    def push(self, o: object = None):
        if o is None:
            self.LOG.error(
                f"A {type(None)} object provided to add to repository"
            )

        idx = []
        if isinstance(o, dict):
            idx = [o.get(self.index_name)]
            o.pop(self.index_name)
            o = [o]
        elif isinstance(o, list):
            idx = [o.get(self.index_name) and o.pop(self.index_name)]
        df = pd.DataFrame(o, index=[idx])
        self._df.append(df)
        return self

    def __dict__(self):
        pass

    def append(self, input_data: DataFrame):
        self.LOG.debug("input dztz = ", extra={"input_data": input_data})
        self.LOG.debug("input dztz = ", extra={"input_data": input_data})
        self._df.append(DataFrame.to_dict(input_data, orient="index"))
        self.LOG.debug("input dztz = ", extra={"self._df": self._df})

    @property
    def dataframe(self):
        return self._df

    def max_by_column(self, column):
        return self._df[column].max()

    def last(self):
        return self._df.iloc[-1]

    def __len__(self):
        """
        Return repository lengh.

        :return: Repository length.
        """
        return len(self._df)



#!/usr/bin/env python3

"""
Utilities for exceptions management.

This file is a part of (denier.io).

(c) 2021, Deepnox SAS.
"""

def raise_if_param_is_none(**kwargs):
    """
    Raise a {TypeError} if provided parameter is None.

    :param name: The parameter name.
    :param value: The value of parameter.
    :return: False if parameter value is not None
    :raise: {TypeError} if value of param is None.
    """
    def _raise_exception(k):
        raise TypeError(f"Parameter is missing: type({k}) is None")

    for k, v in kwargs.items():
        if v is None:
            _raise_exception(k)

    return False





#!/usr/bin/env python3

"""
Helpers to develop metaclass.
"""
from types import FunctionType

def get_attributes_by_type(attrs, type_):
    """
    Returns tye filtered list of user defined attributes from a attributes dictionary passed to the __new__ coro of a metaclass.

    :param attrs: Attributes of a metaclass (last parameters of __new__ coro).
    :return:
    :rtype: dict
    """
    return {k: v for k, v in attrs.items() if isinstance(v, type_) or issubclass(v.__class__, type_)}

def get_user_defined_functions(attrs):
    """
    Returns list of user defined functions from a attributes dictionary passed to the __new__ coro of a metaclass.

    :param attrs: Attributes of a metaclass (last parameters of __new__ coro).
    :rtype: dict
    :return:
    """
    # return get_attributes_by_type(attrs, FunctionType)
    r = {}
    for k, v in attrs.items():
        if isinstance(v, FunctionType):
            if not v.__name__.startswith('__') and not v.__name__.endswith('__'):
                r[k] = v
    return r



#!/usr/bin/env python3

from functools import wraps
from types import FunctionType

from deepnox import loggers
from deepnox.utils.pyutils import copy_func
from deepnox.utils.strings_utils import remove_slash_at_start_and_at_end

LOGGER = loggers.factory(__name__)
""" The module LOGGER. """


def run_as_async_request(coro):
    """
    Run a HttpRequest object as an asynchronous request.
    :param coro:
    :return:
    """
    @wraps(coro)
    async def _impl(self, *method_args, **method_kwargs):
        LOGGER.debug('request._impl wrapper', extra={'method_args': method_args,
                                                     'method_kwargs': method_kwargs})
        req = coro(self, *method_args, **method_kwargs)
        return await self._do_request(**req)
    return _impl

# def to_async(coro):
#     @wraps(coro)
#     def _impl(self, *method_args, **method_kwargs):
#         LOGGER.debug("request._impl wrapper", extra={"method_args": method_args,
#                                                      "method_kwargs": method_kwargs})
#         req = coro(self, *method_args, **method_kwargs)
#
#         return self._do_request(**req)
#
#     return _impl

class ResourceMetaClass(type):
    """
    The metaclass used to create derived {Resource} class.
    """

    LOG = LOGGER.getChild('ResourceMetaClass')
    """ The metaclass LOGGER. """

    _instances = {}
    """ A {dict} containing :class:`deepnox.aiorest.resources.Resource` objects which must be a singleton. """

    def __new__(mcs, clsname, bases, attrs):
        """
        __new__

        :param clsname:
        :param bases:
        :param attrs:
        """
        mcs.LOG.debug(f'__new____new__(mcs, clsname, bases, attrs)', extra={'_mcs': mcs,
                                                                      '_clsname': clsname,
                                                                      '_bases': bases,
                                                                      '_attrs': attrs,})
        print({'_mcs': mcs,
               '_clsname': clsname,
               '_bases': bases,
               '_attrs': attrs,})
        new_attrs = attrs
        new_attrs['_attrs'] = {}
        new_attrs['_request_builders'] = {}
        new_attrs['_endpoints'] = {}

        for k, v in attrs.items():
            if type(v) == str and k in ['name', 'prefix']:
                new_attrs['_attrs'][k] = v
            elif isinstance(v, FunctionType) and not v.__name__.startswith('__') and not v.__name__.endswith('__'):
                new_attrs['_request_builders'][k] = copy_func(v)
                new_attrs['_endpoints'][k] = None  # Will be defined at __call__

            # new_attrs['_request_builders'][k] = v

        return super(ResourceMetaClass, mcs).__new__(mcs, clsname, bases, new_attrs)

    def __call__(self, *args, **kwargs):
        self.LOG.debug('__call__(cls, *args, **kwargs', extra={'_self':self, '_args': args, '_kwargs': kwargs})
        obj = self.__new__(self, *args, **kwargs)
        obj.__init__(*args, **kwargs)
        print('obj.client', obj.client, list(self._endpoints.keys()))
        for request_builder_name in self._endpoints.keys():
            request_builder_fn = self._request_builders.get(request_builder_name)
            # setattr(self, f"_{request_builder_name}_request_builder", request_builder_fn)
            return request_builder_fn(obj, *args, **kwargs)
        return obj




class Resource(object, metaclass=ResourceMetaClass):
    """
    The base class to define {Resource}
    """

    LOG = LOGGER.getChild('Resource')
    """ The LOGGER. """

    def __init__(self, **kwargs):
        self.LOG.debug('__init__(self, **kwargs)', extra={'kwargs': kwargs})
        self.client = kwargs.get('client')

    async def _do_request(self, **req):
        self.LOG.debug('_do_request(self, req)', extra={'req': req})
        x = await self.client._fetch(**req)
        return x

    def absolute_endpoint_url(self, url: str = None):
        """
        Returns absolute URL of endpoint.

        :param url: The relative URL of service.
        :return: The absolute URL of endpoint.
        :rtype: str
        """
        print('1')
        self.LOG.debug('absolute_endpoint_url(self, url: str = None)', extra={'url': url})
        url = url if url is not None and type(url) == str and len(url) > 0 else ''
        print('2')

        r = '/'.join([remove_slash_at_start_and_at_end(str(self.client.base_url)),
         remove_slash_at_start_and_at_end(self._attrs['prefix']),
         remove_slash_at_start_and_at_end(url)])
        return r

    def headers(self, d: dict = None):
        """
        Returns HTTP headers for request.

        :param d: The HTTP headers.
        :type d: dict
        :return: The final HTTP headers (updated using common base headers).
        :rtype: dict
        """
        if d is None:
            d = {}
        d.update(self.client._common_headers)
        if self.bearer_token is not None:
            d.update({'Authorization': 'Bearer {self.bearer_token}'})
        return d



#!/usr/bin/env python3

"""
String utilities.

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""
import re


def remove_slash_at_start_and_at_end(s: str) -> str:
    """
    Remove slash if it's the first or last char.
    Used to normalize URL.

    :param s: String to process.
    :type s: str
    :return: A normalized string without slash
    :rtype: str
    """
    if s is None:
        return None
    r = str(s)
    if len(r) == 0:
        return r
    r = r.strip()
    if r[0] == "/":
        r = r[1:]
    if len(r) > 0 and r[len(r)-1] == "/":
        r = r[:len(r)-1]
    return r


def extract_idp_code(s) -> str:
    """
    Extract an UUID v4 from string.

    :param s: String to process.
    :type s: str
    :return: The extracted UUID.
    :rtype: str
    """
    if not s:
        return
    m = re.findall(r'([a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12})', s)
    if len(m) > 0:
        return m[0]



#!/usr/bin/env python3

"""
Datetime utilities.

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

import arrow

DTKEY_FORMAT = "YYYYMMDDHHmm"

def get_dtkey(dt):
    """
    Returns datetime key used to define period store.
    """
    return arrow.get(dt).format(DTKEY_FORMAT)




#!/usr/bin/env python3
import aiohttp


class BaseAuthorization(object):
    """
    A base class for managing authorizations/permissions.
    """

    def __init__(self, auth_obj: object = None):
        self._auth_obj = auth_obj

    @classmethod
    def from_dict(cls, d: dict) -> object:
        """
        :return: Object as a JSON string.
        """
        if not isinstance(d, dict):
            raise TypeError(f'parameter must be of type: `dict`')
        o = cls()
        for k, v in d.items():
            if k in o.__attrs__:
                setattr(o, k, v)
        return o

class HttpBasicAuth(BaseAuthorization):
    """
    An encapsulation of basic HTTP authentication.
    """

    def __init__(self, username: str = None, password: str = None, encoding: str = 'latin1'):
        self._auth_obj = aiohttp.BasicAuth(username, password, encoding)


#!/usr/bin/env python3

"""
Module: deepnox.aiorest.credentials

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""
import aiohttp

from deepnox.core.enumerations import DeepnoxEnum


class Credentials(object):
    pass


class BasicAuth(Credentials):
    def __init__(self, username: str, password: str, encoding: str = "latin1"):
        super().__init__()
        self._username, self._password, self._encoding = (
            username,
            password,
            encoding,
        )

    def get(self):
        return aiohttp.BasicAuth(
            self._username, self._password, self._encoding
        )


class AuthorizationType(DeepnoxEnum):
    """
    The different supported credentials type.
    """

    NO_AUTH = 0
    """ No authentication. """

    API_KEY = 1
    """ The authentication based on a bearer token. """

    BEARER_TOKEN = 2
    """ The authentication based on a bearer token. """

    BASIC_AUTH = BasicAuth
    """ The HTTP basic authentication. """

    def __instancecheck__(self, instance):
        if issubclass(self.value, instance.__class__):
            return True
        return False



#!/usr/bin/env python3

import functools
import types


def get_attributes(obj):
    return {attr_name: getattr(obj, attr_name) for attr_name in dir(obj)
                      if callable(getattr(obj, attr_name))}

def copy_func(f):
    """
    Copy a function.

    Based on:

    - http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)
    - [How to create a copy of a python function](https://stackoverflow.com/a/13503277)
    """
    g = types.FunctionType(f.__code__, f.__globals__, name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return g




def percentage(part, whole):
    """
    Compute a percentage.

    :param part: The part.
    :type part: float
    :param whole: The whole.
    :type whole: float
    :return: The computed percentage.
    """
    return 100 * float(part) / float(whole)



#!/usr/bin/env python3

"""
A JSON serializer.

Package: deepnox.serializers.json_serializers

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""
import datetime
import json
import sys

from deepnox.core.enumerations import DeepnoxEnum
from deepnox.serializers.base_serializer import BaseSerializer
from deepnox.third import arrow


def is_json(s):
    """
    Return True if provided string is a JSON object.
    :param s: The string to test.
    :return: True if JSON. False else.
    """
    try:
        json.loads(s)
    except ValueError:
        return False
    return True


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return arrow.get(obj).isoformat()
        elif isinstance(obj, arrow.Arrow):
            return obj.isoformat()
        elif isinstance(obj, (type, DeepnoxEnum)):
            return str(obj)
        else:
            try:
                return json.JSONEncoder.default(self, obj)
            except TypeError as e:
                return str(obj)


class JsonSerializer(BaseSerializer):
    """
    JSON record_serializer.
    """

    def __init__(self, encoder: json.JSONEncoder = None):
        """
        Crate a new JSON record_serializer.
        """
        if encoder is None:
            encoder = ComplexEncoder
        super().__init__(name="json", encoder=encoder)

    def dump(self, o: object):
        """
        Serialize an object as JSON string.
        :param o: The object to serialize.
        :type o: object
        :return: The serializing object as string.
        :rtype: str
        """
        return json.dumps(o, cls=self.encoder)

    def load(cls, s: str) -> object:
        """
        Unserialize a string from JSON.
        :param s: The string to unserialize.
        :type s: str
        :return: The unserializing string object as object.
        :rtype: str
        """
        return json.loads(s)





class SingletonMeta(type):
    """
    A metaclass to use other classes as singleton.
    :see: https://www.datacamp.com/community/tutorials/python-metaclasses
    """

    _instances = {}
    """ A {dict} containing classes which must be a singleton. """

    def __call__(cls, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta,cls).__call__(*args, **kwargs)
        return cls._instances[cls]



#!/usr/bin/env python3

"""
Package: deepnox.loggers

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

from . import defaults

import logging
import logging.config
import os

__import__("pkg_resources").declare_namespace(__name__)

# @see: https://gist.github.com/pmav99/49c01313db33f3453b22

# A regarder absolument:
# * https://towardsdatascience.com/8-advanced-python-logging-features-that-you-shouldnt-miss-a68a5ef1b62d
# * https://gist.github.com/mariocj89/73824162a3e35d50db8e758a42e39aab
# * Terminal colorized: https://gist.github.com/exhuma/8147910


# Define MESSAGE log level
AUDIT = 25

# "Register" new loggin level
logging.addLevelName(AUDIT, "AUDIT")  # addLevelName(25, 'MESSAGE')

# Verify
assert logging.getLevelName(AUDIT) == "AUDIT"


def audit(self, msg, *args, **kwargs):
    if self.isEnabledFor(AUDIT):
        self._log(AUDIT, msg, args, **kwargs)


logging.Logger.audit = audit

DEFAULT_LOGGER_NAME = "root" or os.environ.get("LG_LOGGER_NAME")
"""The default loggers name.
"""


def setup(config: dict = defaults.DEFAULT_LOG_SETTINGS):
    """Setup loggers from provided configuration as dict.

    :param config: The wished configuration for loggers.
    """
    logging.config.dictConfig(config=config)


def update_defaults_setup(config: dict = defaults.DEFAULT_LOG_SETTINGS):
    """Setup loggers from provided configuration as dict.

    :param config: The wished configuration for loggers.
    """
    config = {**defaults.DEFAULT_LOG_SETTINGS, **config}
    logging.config.dictConfig(config=config)


def factory(name: str = DEFAULT_LOGGER_NAME, suffix: str = None):
    """A simple loggers factory.

    :param name: The loggers name.
    :param suffix: Optional suffix loggers
    :return: An instance of official Python {logging.Logger}
    """
    if isinstance(suffix, str) and len(suffix) > 0:
        try:
            return logging.getLogger(name).getChild(suffix=suffix)
        except Exception:
            return logging.getLogger(name)
    return logging.getLogger(name)


def auditor(project_name: str):
    return factory(f"loadguard.tests.load.metrics.{project_name}")



#!/usr/bin/env python3

"""
A YAML serializer.

Package: deepnox.serializers.yaml_serializers

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""
from typing import Any

import yaml

from deepnox.serializers.base_serializer import BaseSerializer


def is_yaml(s):
    """
    Return True if provided string is a yaml object.
    :param s: The string to test.
    :return: True if yaml. False else.
    """
    try:
        yaml.load(s, Loader=yaml.FullLoader())
    except ValueError:
        return False
    return True


class YamlSerializer(BaseSerializer):
    """
    yaml record_serializer.
    """

    def __init__(self, encoder: Any = None):
        """
        Crate a new yaml record_serializer.
        """
        super().__init__(name="yaml", encoder=encoder)

    def dump(self, o: object):
        """
        Serialize an object as yaml string.
        :param o: The object to serialize.
        :type o: object
        :return: The serializing object as string.
        :rtype: str
        """
        return yaml.dump(o)

    def load(self, s: str) -> object:
        """
        Unserialize a string from yaml.
        :param s: The string to unserialize.
        :type s: str
        :return: The unserializing string object as object.
        :rtype: str
        """
        return yaml.load(s, Loader=yaml.FullLoader)



#!/usr/bin/env python3

"""

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.

"""

from deepnox import loggers
from deepnox.core.enumerations import UniqueEnum

LOGGER = loggers.factory(__name__)
""" The main LOGGER. """


class ContentType(UniqueEnum):
    """
    Non-exhaustive enumeration of content types.
    """

    APPLICATION__JSON = "application/json"
    """ The `application/json` content type. """

    APPLICATION__X_WWW_FORM_URLENCODED = "application/x-www-form-urlencoded"
    """ The `application/x-www-form-urlencoded` content type. """

    TEXT__PLAIN = "text/plain"
    """ The `text/plain` content type. """

    TEXT__HTML = "text/html"
    """ The `text/html` content type. """

    def __str__(self):
        return self.value



#!/usr/bin/env python3

"""
Package: deepnox.third

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

__import__('pkg_resources').declare_namespace(__name__)


import aiohttp
import arrow
import numpy
import pandas
import pydantic
import requests
import shortuuid
import yaml


aiohttp = aiohttp
arrow = arrow
# jinja2 = __import__("jinja2")
numpy = numpy
pandas = pandas
pydantic = pydantic
requests = requests
shortuuid = shortuuid
yaml = yaml





#!/usr/bin/env python3

"""
The base class for serializing services.

Package: deepnox.serializers.base_serializers

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""
from typing import Any


class BaseSerializer(object):
    """
    A base class for record_serializer.
    """

    def __init__(self, name: str, encoder: Any = None):
        """
        Create a new record_serializer.
        """
        self.name = name
        self.encoder = encoder

    def dump(self, o: object):
        """
        Serialize an object as JSON string.
        :param s: The object to serialize.
        :type s: str
        :return: The serializing object as string.
        :rtype: str
        """
        raise NotImplementedError("`serialize` is not implemented")

    def load(self, s: str) -> object:
        """
        Unserialize a string from JSON.
        :param s: The string to unserialize.
        :type s: str
        :return: The unserializing string object as object.
        :rtype: str
        """
        raise NotImplementedError("`deserialize` is not implemented")



#!/usr/bin/env python3

"""
This module offers features facilitating configuration management.

Module: deepnox.settings.base

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

import os

import yaml

from deepnox import loggers
from deepnox.core.base import DeepnoxDict
from deepnox.utils.maps import UpperMap

LOGGER = loggers.factory(__name__)
""" The module LOGGER. """


class NotFoundConfigurationError(IOError):
    """
    An exception if configuration can't be found.
    """

    def __init__(self, filename: str, message="{filename} is not found"):
        """
        Create a new instance of {NotFoundConfigurationError}
        :param filename:
        :param message:
        """
        self.filename: str = filename
        self.message: str = message.format(filename=self.filename)
        super().__init__(self.message)


def read_yaml_file(filename: str):
    """
    Open and read a YAML file.
    :param filename: The filename to deserialize.
    :return: The YAML object.
    """
    with open(filename, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            LOGGER.error(
                f"Error while reading yaml file: {filename}", exc_info=exc
            )


class SettingsReader(DeepnoxDict):
    """
    A class to manage settings.
    """

    def __init__(self, filename: str = None):
        """
        Create a new settings object from provided configuration (a YAML file).
        :param filename: The configuration filename.
        """
        if filename is None or os.path.isfile(filename) is False:
            raise NotFoundConfigurationError(filename=filename)
        self._settings = UpperMap(read_yaml_file(filename))

    def __iter__(self):
        for attr, value in self._settings.items():
            yield attr, value

    def __getattr__(self, item):
        if item in self._settings.keys():
            return self._settings.get(item)
        return super().__getattr__(item)





#!/usr/bin/env python3

"""

Package: deepnox.

This file is a part of LoadGuard Runner.

(c) 2021, Deepnox SAS.
"""

from enum import unique

from deepnox.core.enumerations import DeepnoxEnum


@unique
class Scheme(DeepnoxEnum):
    """ Unexhaustive enumeration of schemes.
    """

    FILE = 'file'
    """ The file scheme. """

    HTTP = 'network'
    """ The HTTP scheme. """

    HTTPS = 'https'
    """ The HTTPS scheme. """

    FTP = 'ftp'
    """ The FTP scheme. """

    FTPS = 'ftps'
    """ The FTPS scheme. """

    WEBSOCKET = 'ws'
    """ The websocket scheme. """

    SSH = 'ssh'
    """ The SSH scheme. """

    GIT = 'git'
    """ The git scheme. """

    def __str__(self):
        return self.value







#!/usr/bin/env python3

"""
Auditing utilities.

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

import asyncio
import time
import logging
from functools import wraps


from deepnox import loggers

LOGGER = loggers.factory(__name__)
""" The main LOGGER. """


def timeit(logger: logging.Logger, trace_params: bool = False):
    """
    A decorator to trace time execution.

    :param logger: The LOGGER.
    :param trace_params: Should I trace parameters?
    """

    def _timeit(f):
        @wraps(f)
        def wrap(*args, **kwargs):
            ts = time.monotonic()
            result = f(*args, **kwargs)
            te = time.monotonic()
            extra_trace = {
                'func_name': f.__name__,
                'execution_time': (te - ts),
            }
            if trace_params is True:
                extra_trace.update({'parameters': {'args': args, 'kwargs': kwargs}})
            logger.info(f'Execution time: {f.__name__}', extra=extra_trace)
            return result

        return wrap

    return _timeit


def async_timeit(func):
    async def process(func, *args, **params):
        if asyncio.iscoroutinefunction(func):
            print('this function is a coroutine: {}'.format(func.__name__))
            return await func(*args, **params)
        else:
            print('this is not a coroutine')
            return func(*args, **params)

    async def helper(*args, **params):
        print('{}.time'.format(func.__name__))
        start = time.time()
        result = await process(func, *args, **params)

        # Test normal function route...
        # result = await process(lambda *a, **p: print(*a, **p), *args, **params)

        print('>>>', time.time() - start)
        return result

    return helper


async def compute(x, y):
    print('Compute %s + %s ...' % (x, y))
    await asyncio.sleep(1.0)  # asyncio.sleep is also a coroutine
    return x + y


@timeit
async def print_sum(x, y):
    result = await compute(x, y)
    print('%s + %s = %s' % (x, y, result))

# loop = asyncio.get_event_loop()
# loop.run_until_complete(print_sum(1, 2))
# loop.close()



import asyncio
import json
import logging
import sys
import time

import aiohttp
from aiohttp import TCPConnector

from deepnox import loggers
from deepnox.aiorest.service import Resource
from deepnox.auth.base import BaseAuthorization
from deepnox.network.http import HttpMethod
from deepnox.network.http import HttpResponse
from deepnox.network.urls import Url
from deepnox.serializers.json_serializer import is_json
from deepnox.utils.metacls import get_attributes_by_type

LOGGER = loggers.factory(__name__)
""" The main LOGGER. """


class RestClientMetaclass(type):
    """
    The metaclass to create a derived class of {AioRestClient}.
    """

    LOG = LOGGER.getChild('RestClientMetaclass')
    """ The metaclass LOGGER. """

    def __new__(mcs, clsname, bases, attrs):
        """
        The `__new__` method.

        :param clsname: The class name.
        :param bases: The base classes.
        :param attrs: The attributes of the created class.
        """
        # RestClientMetaclass.LOG.debug("__new__(mcs, clsname, bases, attrs)",
        #                               extra={"mcs": mcs, "clsname": clsname, "bases": bases, "attrs": attrs})
        new_attrs = attrs  # Copy coriginal attributes to the target containing new attributes.
        new_attrs["_services"] = {}  #

        # For each web service found in the list, store name and value of the service in a dictionary of the class
        for k, v in get_attributes_by_type(attrs, Resource).items():
            new_attrs["_services"][k] = v

        # Returns a class with new attributes and/or functions.
        return super(RestClientMetaclass, mcs).__new__(mcs, clsname, bases, new_attrs)

    def __call__(self, *args, **kwargs):
        """
        The __call__ method.

        :param args: The arguments.
        :param kwargs: The kwargs.
        :return: ? (sic)

        :todo: confirm utility ?
        """
        self.LOG.debug("__call__(self, *args, **kwargs)", extra={"_args": args, "_kwargs": kwargs})
        obj = self.__new__(self, *args, **kwargs)
        obj.__init__(*args, **kwargs)
        for k, services_obj in obj._services.items():
            self.LOG.debug(f"__call__(self, *args, **kwargs): inject {obj} in {services_obj}", extra={
                "k": k,
                "obj": obj,
                "services_obj": services_obj,
            })
            setattr(services_obj, "client", obj)
        return obj


class AioRestClient(object, metaclass=RestClientMetaclass):
    LOG = LOGGER.getChild('AioRestClient')
    """ The class LOGGER. """

    def __init__(self,
                 loop: asyncio.AbstractEventLoop = None,
                 base_url: Url = None,
                 common_headers: str = None,
                 verify_ssl: bool = False,
                 auth: BaseAuthorization = None,
                 requote_redirect_url: bool = False,
                 auditor_logger: logging.Logger = None,
                 ):
        """
        Create a new instance of {AioRestClient}.

        :param loop: The event loop.
        :type loop: asyncio.AbstractEventLoop
        :param base_url: The base URL of the group of services.
        :type base_url: Url
        :param common_headers: HTTP headers used when invoking any API endpoint.
        :param verify_ssl: Boolean to set if TLS/SSL certificates must be verified.
        :type verify_ssl: bool
        """
        self.loop: asyncio.AbstractEventLoop = loop or asyncio.get_event_loop()
        self.base_url: Url = base_url
        self._common_headers = common_headers
        self.history = []
        self._connector = self._create_connector(verify_ssl=verify_ssl)
        self._auth = auth
        self._session = None
        self._requote_redirect_url = requote_redirect_url
        self.AUDITOR = auditor_logger or self.LOG


    def _create_connector(self, **kwargs) -> TCPConnector:
        """
        Return a connector without TLS verification by default.

        :param kwargs: The arguments to pass when creating the TCP connector.
        :return:
        """
        kwargs["verify_ssl"] = kwargs.get("verify_ssl") or False
        # kwargs["raise_for_status"] = kwargs.get("raise_for_status") or True
        if kwargs.get("authentication"):
            kwargs["auth"]

        return aiohttp.TCPConnector(**kwargs)

    async def _fetch(self, **kwargs):
        """
        Send an asynchronous HTTP request.

        :return:
        """
        url = kwargs.get("url")
        method = kwargs.get('method') or HttpMethod.GET
        basic_auth = kwargs.get('basic_auth')
        for item in ["method", "url", "basic_auth"]:
            if item in kwargs.keys():
                del kwargs[item]

        self.LOG.debug(f"_request(self, **kwargs)", extra={"url": url, "method": method})

        start_ts = time.monotonic()
        http_response = HttpResponse()
        try:
            async with self.session(basic_auth=basic_auth).request(method=str(method), url=url, **kwargs) as resp:
                self.LOG.debug(f'resp.status: {resp.status}')
                http_response.elapsed_time = int((time.monotonic() - start_ts) * 1000)
                http_response.text = await resp.text()
                if is_json(http_response.text) is True:
                    http_response.json = json.loads(http_response.text)
                    delattr(http_response, "text")
                http_response.status_code = resp.status
                self.AUDITOR.audit(f'Request: {url}', extra={
                    'request': {
                        "url": url,
                        "method": str(method),
                        "_kwargs": kwargs,
                    },
                    'response': http_response.to_dict(),
                })
                self.history.append(http_response)
                return http_response.to_dict()
        except aiohttp.ClientConnectorError as e:
            self.LOG.error(f'Connection error: {e}', exc_info=e)
        except:
            e = sys.exc_info()[0]
            self.LOG.error(f'Uncaught error: {e}', exc_info=e)

    async def get(self, *args, **kwargs):
        """
        Send an asynchronous HTTP request using the GET method.

        :param args:
        :param kwargs:
        :return:
        """
        self.LOG.debug(f'get(args={args}, kwargs={kwargs})')
        kwargs['method'] = HttpMethod.GET
        await self._fetch(**kwargs)

    async def post(self, *args, **kwargs):
        """
        Send an asynchronous HTTP request using the POST method.

        :param args:
        :param kwargs:
        :return:
        """
        self.LOG.debug(f'post(args={args}, kwargs={kwargs})')
        kwargs['method'] = HttpMethod.POST
        return await self._fetch(**kwargs)

    async def put(self, *args, **kwargs):
        """
        Send an asynchronous HTTP request using the PUT method.

        :param args:
        :param kwargs:
        :return:
        """
        self.LOG.debug(f'put(args={args}, kwargs={kwargs})')
        kwargs['method'] = HttpMethod.PUT
        return await self._fetch(**kwargs)

    async def patch(self, *args, **kwargs):
        """
        Send an asynchronous HTTP request using the PATCH method.

        :param args:
        :param kwargs:
        :return:
        """
        self.LOG.debug(f'patch(args={args}, kwargs={kwargs})')
        kwargs['method'] = HttpMethod.PATCH
        return await self._fetch(**kwargs)

    async def delete(self, *args, **kwargs):
        """
        Send an asynchronous HTTP request using the DELETE method.

        :param args:
        :param kwargs:
        :return:
        """
        self.LOG.debug(f'delete(args={args}, kwargs={kwargs})')
        kwargs['method'] = HttpMethod.DELETE
        return await self._fetch(**kwargs)

    async def options(self, *args, **kwargs):
        """
        Send an asynchronous HTTP request using the OPTIONS method.

        :param args:
        :param kwargs:
        :return:
        """
        self.LOG.debug(f'basic_auth(args={args}, kwargs={kwargs})')
        kwargs['method'] = HttpMethod.OPTIONS
        return await self._fetch(**kwargs)

    def __del__(self):
        self.loop.run_until_complete(self._connector.close())



#!/usr/bin/env python3

"""
Module: deepnox.network.schemes

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""


from deepnox.core.enumerations import UniqueEnum


class Scheme(UniqueEnum):
    """
    Unexhaustive enumeration of schemes.
    """

    FILE = "file"
    """ The file scheme. """

    HTTP = "network"
    """ The HTTP scheme. """

    HTTPS = "https"
    """ The HTTPS scheme. """

    FTP = "ftp"
    """ The FTP scheme. """

    FTPS = "ftps"
    """ The FTPS scheme. """

    WEBSOCKET = "ws"
    """ The websocket scheme. """

    SSH = "ssh"
    """ The SSH scheme. """

    GIT = "git"
    """ The git scheme. """

    def __str__(self):
        return self.value



#!/usr/bin/env python3

"""
Defaults settings used for logging.

Module: deepnox.loggers.defaults

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

import sys

from deepnox.loggers.formatters.json_formatter import JsonFormatter
from deepnox.loggers.formatters.result_formatter import ResultFormatter
from deepnox.loggers.formatters.yaml_formatter import YamlFormatter

DEFAULT_LOG_SETTINGS = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "json_formatter": {
            "()": JsonFormatter,
        },
        "yaml_formatter": {
            "()": YamlFormatter,
        },
        "result_formatter": {
            "()": ResultFormatter,
        },
    },
    "handlers": {
        "default": {
            "class": "logging.StreamHandler",
            "formatter": "result_formatter",
            "level": "DEBUG",
            "stream": sys.stdout,
        }
    },
    "loggers": {
        "": {  # root loggers
            "handlers": ["default"],
            "level": "DEBUG",
            "propagate": True,
        },
        "__main__": {  # if __name__ == '__main__'
            "handlers": ["default"],
            "level": "DEBUG",
            "propagate": False,
        },
        "deepnox": {
            "handlers": ["default"],
            "level": "DEBUG",
            "propagate": False,
        },
        "loadguard": {
            "handlers": ["default"],
            "level": "DEBUG",
            "propagate": False,
        },
        "sncf.transilien": {
            "handlers": ["default"],
            "level": "DEBUG",
            "propagate": False,
        },
    },
    "root": {"handlers": ["default"], "level": "INFO", "propagate": False},
}



#!/usr/bin/env python3

"""
The setup script.

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.

"""

import sys
from types import FunctionType
from distutils.core import Command

__name__ = "wipbox"
""" The project name. """

__version__ = "0.0.9"
""" The current version (work in progress). """


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# Cf. https://stackoverflow.com/a/23265673
try:
    from pypandoc import convert

    read_md: FunctionType = lambda f: convert(f, 'rst')
except ImportError:
    print("Warning: pypandoc module not found")
    read_md: FunctionType = lambda f: open(f, 'r').read()



with open("requirements.txt") as f:
    requires = f.readlines()
if sys.version_info[0] < 3 and sys.version_info[1] <= 6:
    raise Exception
else:
    tests_require = []
    test_command = [
        sys.executable,
        "-m",
        "pytest",
    ]  # Run unit tests
    coverage_command = [
        "coverage",
        "run",
        "-m",
        "pytest",
        "test/"
        "&&",
        "coverage",
        "report",
        "-m"
    ]  # Compute coverage
    coverage_xml = [
        "coverage",
        "xml",
    ]  # Export coverage as XML
    coverage_upload = [
        "codacy-coverage-reporter",
        "report",
    ]  # Upload to Codacity


class RunUnitTests(Command):
    """Run unit tests."""

    user_options = []
    description = __doc__[1:]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess

        errno = subprocess.call(test_command)
        raise SystemExit(errno)


class RunTestsCoverage(Command):
    """Run unit tests and report on code coverage using the 'coverage' tool."""

    user_options = []
    description = __doc__[1:]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess

        errno = subprocess.call(coverage_command)
        if errno == 0:
            subprocess.call(["coverage", "xml"])
        raise SystemExit(errno)


class UploadCoverage(Command):
    """Upload code coverage to Codacity."""

    user_options = []
    description = __doc__[1:]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess

        errno = subprocess.call(coverage_upload)
        if errno != 0:
            raise SystemExit(errno)


setup(
    name=__name__,
    version=f"{__version__}",
    description="A set of modern Python libraries under development to simplify the execution of reusable routines by different projects.",
    license="Copyrighted 2021-2022, Deepnox SAS",
    author="The Deepnox team",
    author_email="contact@deepnox.io",
    url="https://github.com/deepnox-io/pythpn-wipbox",
    download_url=f"https://github.com/deepnox-io/pythpn-wipbox/{__version__}.tar.gz",
    long_description=read_md("README.md"),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
    ],
    keywords=[
        "deepnox",
        "wipbox",
        "utilities",
        "tools",
        "various",
        "work in progress",
        "draft",
    ],
    packages=[
        "deepnox",
        "deepnox.aiobox",
        "deepnox.aiorest",
        "deepnox.auth",
        "deepnox.clients",
        "deepnox.core",
        "deepnox.files",
        "deepnox.helpers",
        "deepnox.loggers",
        "deepnox.models",
        "deepnox.network",
        "deepnox.patterns",
        "deepnox.repositories",
        "deepnox.serializers",
        "deepnox.settings",
        "deepnox.third",
        "deepnox.utils",
    ],
    package_dir={
        "deepnox": "src/deepnox",
        "deepnox.aiobox": "src/deepnox/aiobox",
        "deepnox.aiorest": "src/deepnox/aiorest",
        "deepnox.auth": "src/deepnox/auth",
        "deepnox.clients": "src/deepnox/clients",
        "deepnox.core": "src/deepnox/clients",
        "deepnox.files": "src/deepnox/files",
        "deepnox.helpers": "src/deepnox/clients",
        "deepnox.loggers": "src/deepnox/loggers",
        "deepnox.models": "src/deepnox/models",
        "deepnox.network": "src/deepnox/network",
        "deepnox.patterns": "src/deepnox/patterns",
        "deepnox.repositories": "src/deepnox/repositories",
        "deepnox.serializers": "src/deepnox/serializers",
        "deepnox.settings": "src/deepnox/settings",
        "deepnox.third": "src/deepnox/third",
        "deepnox.utils": "src/deepnox/utils",
    },
    scripts=[],
    install_requires=requires,
    tests_require=tests_require,
    cmdclass={
        "test": RunUnitTests,
        'coverage': RunTestsCoverage,
        # 'upload': UploadCoverage
    },
)



#!/usr/bin/env python3

"""
A serializer for native Python pickle format.

Package: deepnox.serializers.pickle_serializers

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

from deepnox.serializers.base_serializer import BaseSerializer


class PickleSerializer(BaseSerializer):
    """
    yaml record_serializer.
    """

    def __init__(self):
        """
        Crate a new yaml record_serializer.
        """
        super().__init__(name="pickle")

    def dump(self, o: object):
        """
        Serialize an object as yaml string.
        :param o: The object to serialize.
        :type o: object
        :return: The serializing object as string.
        :rtype: str
        """
        pass

    def load(self, s: str):
        pass



#!/usr/bin/env python3

"""
This package provides simple routines for defining models or schemas.

Package: deepnox.models

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

__import__("pkg_resources").declare_namespace(__name__)
from typing import Union, Any, Dict

import pydantic


class ExtendedBaseModel(pydantic.BaseModel):
    """
    An extended Pydantic model to support definition of properties.

    Resources:

    - https://github.com/samuelcolvin/pydantic/issues/935#issuecomment-554378904
    - https://stackoverflow.com/questions/63264888/pydantic-using-property-getter-decorator-for-a-field-with-an-alias

    """

    @classmethod
    def get_properties(cls):
        return [prop for prop in cls.__dict__ if isinstance(cls.__dict__[prop], property)]

    def dict(
            self,
            *,
            include: Union['AbstractSetIntStr', 'DictIntStrAny'] = None,
            exclude: Union['AbstractSetIntStr', 'DictIntStrAny'] = None,
            by_alias: bool = False,
            skip_defaults: bool = None,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = True,
    ) -> Dict[str, Any]:
        """Override the dict function to include our properties"""
        attribs = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none
        )
        props = self.get_properties()

        # Include and exclude properties
        if include:
            props = [prop for prop in props if prop in include]
        if exclude:
            props = [prop for prop in props if prop not in exclude]
        if exclude_none:
            props = [prop for prop in props if getattr(self, prop) is not None]

        # Update the attribute dict with the properties
        if props:
            attribs.update({prop: getattr(self, prop) for prop in props})
        return attribs



#!/usr/bin/env python3

"""
Module: deepnox.aiorest.client

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""

import asyncio
import inspect
import sys
from typing import Dict, Tuple, Type, Union

from aiohttp import ClientSession

from deepnox import loggers
from deepnox.aiorest.credentials import AuthorizationType, BasicAuth, Credentials
from deepnox.aiorest.resources import Resource
from deepnox.network import Scheme
from deepnox.network.urls import Url
from deepnox.utils.maps import CannotConvertToMapTypeError, Map, to_map
from deepnox.utils.metacls import get_attributes_by_type

LOGGER = loggers.factory(__name__)
""" The module LOGGER. """

loggers.setup()


class SessionArgsBuilder(object):
    def __init__(self, o):
        self._obj = o
        self._args = {'loop': o.loop}
        self.add_basic_auth()

    def get(self):
        return self._args

    def add_basic_auth(self):
        if (
                isinstance(self._obj.credentials, AuthorizationType.BASIC_AUTH)
                is True
        ):
            self._args['auth'] = self._obj.credentials.get()
        return self


class BaseRestClient(object):
    """
    The base class from which the :class:`deepnox.aiorest.client.RestClient` class.
    """

    def __init__(
            self,
            base_url: Url,
            loop: asyncio.AbstractEventLoop = None,
            credentials: Tuple[Credentials] = None,
            connector_options: Type[Union[Dict, Map]] = None,
            base_http_headers: Type[Union[Dict, Map]] = None,
    ):
        for k in self._resources.keys():
            self._resources[k].client = self

        self.base_url = base_url
        self.loop = loop or asyncio.get_event_loop()
        self.credentials = credentials
        self.connector = to_map(connector_options)
        self.base_http_headers = to_map(base_http_headers)
        self._session = None

    async def session(self, **kwargs):
        self._session = ClientSession(**SessionArgsBuilder(self).get())
        return self._session

    def __getattribute__(self, item: str) -> Resource:
        if isinstance(item, Resource):
            return self._resources[item]
        return super().__getattribute__(item)


class RestClientMetaClass(type):
    """
    The metaclass to simplify declaration of resources list.
    """

    LOG = LOGGER.getChild('RestClientMetaclass')
    """ The metaclass LOGGER. """

    def __init__(self, what, bases=None, dict=None):
        print('dict', dict)
        for k in inspect.getmembers(bases[0]):
            print('member =', k)


    def __new__(mcs, clsname, bases, attrs):
        """
        The `__new__` method.

        :param clsname: The class name.
        :param bases: The base classes.
        :param attrs: The attributes of the created class.
        """
        # RestClientMetaclass.LOG.debug("__new__(mcs, clsname, bases, attrs)",
        #                               extra={"mcs": mcs, "clsname": clsname, "bases": bases, "attrs": attrs})
        new_attrs = attrs  # Copy coriginal attributes to the target containing new attributes.
        print('clsname, bases, attrs', clsname, bases, attrs.items())


        new_attrs['_resources'] = {}  #
        for k, v in attrs.items():
            print('k, v == ', k, v)
        return super().__new__(
            mcs, clsname, bases, new_attrs
        )




class RestClient(BaseRestClient, metaclass=RestClientMetaClass):

    def __init__(
            self,
            base_url: Url,
            loop: asyncio.AbstractEventLoop = None,
            credentials: Tuple[Credentials] = None,
            connector_options: Type[Union[Dict, Map]] = None,
            base_http_headers: Type[Union[Dict, Map]] = None,
    ):
        super().__init__(base_url,
                         loop,
                         credentials,
                         connector_options,
                         base_http_headers,
                         )


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    base_url = Url(scheme=Scheme.HTTP, hostname='test.com')
    credentials = BasicAuth('username', 'password')
    client = BaseRestClient(
        loop=loop, base_url=base_url, credentials=credentials
    )

    session = loop.run_until_complete(client.session())
    print(session)



#!/usr/bin/env python3

"""
Module: deepnox.tests.network.http

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""
import json
from dataclasses import Field
from datetime import datetime
from enum import EnumMeta, unique
from typing import Dict, Optional, Any, Union
from uuid import UUID, uuid4

from pydantic import validator

from deepnox.core.enumerations import DeepnoxEnum
from deepnox.models import ExtendedBaseModel
from deepnox.network.urls import Url
from deepnox.serializers.json_serializer import is_json
from deepnox.third import arrow, pydantic, yaml


class HttpMethodMetaClass(EnumMeta):
    def __call__(cls, value, *args, **kwargs):
        if isinstance(value, str):
            value = value.lower()
            return super().__call__(value, *args, **kwargs)


@unique
class HttpMethod(DeepnoxEnum, metaclass=HttpMethodMetaClass):
    """
    Enumeration of HTTP methods.
    """

    GET = 'get'
    """ The HTTP GET coro. """

    POST = 'post'
    """ The HTTP POST coro. """

    PUT = 'put'
    """ The HTTP PUT coro. """

    DELETE = 'delete'
    """ The HTTP DELETE coro. """

    OPTIONS = 'options'
    """ The HTTP OPTIONS coro. """

    PATCH = 'patch'
    """ The HTTP PATCH coro. """

    HEAD = 'head'
    """ The HTTP HEAD coro. """

    @classmethod
    def get(cls, s):
        return getattr(cls, s.upper())


class HttpRequestPayload(ExtendedBaseModel, extra=pydantic.Extra.forbid, orm_mode=True):
    """
    Payload of a HTTP request.

    """

    params: Optional[Dict] = None
    """ The querystring parameters. """

    data: Optional[Union[str, Dict]] = None
    """ The raw body passed to request. """


class HttpRequest(ExtendedBaseModel, extra=pydantic.Extra.forbid, orm_mode=True):
    """
    An HTTP request.

    """

    method: HttpMethod = HttpMethod.GET
    """ The HTTP method to use. """

    url: Optional[Union[Url, Dict]] = None
    """ The targeted url."""

    headers: Optional[Dict] = None
    """ The HTTP request headers."""

    payload: Optional[HttpRequestPayload] = None
    """ The HTTP payload. """

    params: Optional[Dict] = None
    """ The parameters to send. """

    data: Optional[str] = None
    """ The data to send. """

    json_data: Optional[Dict] = None
    """ The JSON to send. """

    authorization: Optional[str] = None
    """
    The provided authorization.
    
    :todo: Implements authorization.
    """

    start_at: Optional[datetime] = None
    """ Datetime starting request. """

    end_at: Optional[datetime] = None
    """ Datetime ending request. """

    body: Optional[str]

    @validator('url', pre=True, always=True)
    def url_autoconvert(cls, v):
        if isinstance(v, Url):
            return v
        if isinstance(v, dict):
            return Url(**v)
        raise TypeError

    @validator('payload', pre=True, always=True)
    def payload_autoconvert(cls, v):
        if isinstance(v, HttpRequestPayload):
            return v
        if isinstance(v, dict):
            return HttpRequestPayload(**v)

    @property
    def size(self) -> int:
        if self.body is not None:
            return len(self.body)
        return 0

    def dict(
            self,
            **kwargs
    ) -> Dict[str, Any]:
        kwargs.update({"exclude_none": True})
        return super().dict(**kwargs)


class HttpGetRequest(HttpRequest):
    """
    The GET method for HTTP protocol.
    """

    method: HttpMethod = HttpMethod.GET
    """ The HTTP method to use. """


class HttpPostRequest(HttpRequest):
    """
    The POST method for HTTP protocol.
    """

    method: HttpMethod = HttpMethod.POST
    """ The HTTP method to use. """


class HttpPutRequest(HttpRequest):
    """
    The PUT method for HTTP protocol.
    """

    method: HttpMethod = HttpMethod.PUT
    """ The HTTP method to use. """


class HttpPatchRequest(HttpRequest):
    """
    The PATCH method for HTTP protocol.
    """

    method: HttpMethod = HttpMethod.PATCH
    """ The HTTP method to use. """


class HttpDeleteRequest(HttpRequest):
    """
    The DELETE method for HTTP protocol.
    """

    method: HttpMethod = HttpMethod.DELETE
    """ The HTTP method to use. """


class HttpOptionsRequest(HttpRequest):
    """
    The OPTIONS method for HTTP protocol.
    """

    method: HttpMethod = HttpMethod.OPTIONS
    """ The HTTP method to use. """


class HttpResponse(ExtendedBaseModel, extra=pydantic.Extra.forbid, orm_mode=True):
    """
    A response summary for HTTP protocol.
    """

    status_code: Optional[int]
    """ The HTTP status code. """

    size: Optional[int]
    """ The HTTP response size (in bytes). """

    headers: Optional[Dict]
    """ The HTTP headers as a key/value dictionary. """

    text: Optional[str]
    """ The text of the response. """

    _text_copy: Optional[str]
    """ A backup of the response text. """

    end_at: Optional[datetime]
    """ The response has been finished at... """

    elapsed_time: Optional[float]
    """ Elapsed time between receiving response and starting request. """


class HttpHit(ExtendedBaseModel):
    """
    A HTTP hit.
    """

    # request_id: UUID = Field(default_factory=uuid4)
    """ The unique request identifier. """

    status_code: Optional[int]
    """ The HTTP status code. """

    url: Optional[Url]
    """ The targeted URL. """

    method: Optional[HttpMethod]
    """ The HTTP request method. """

    start_at: Optional[datetime]
    """ Start datetime of request process. """

    end_at: Optional[datetime]
    """ End datetime of request process. """

    request: Optional[HttpRequest]
    """ The HTTP request. """

    response: Optional[HttpResponse]
    """ The HTTP response. """

    error: Optional[str]
    """ Error description if occurs. """

    @property
    def elapsed_time(self):
        if self.end_at is not None and self.start_at is not None:
            return (arrow.get(self.end_at).timestamp() - arrow.get(self.end_at).timestamp()) * 1000
        return

    class Config:
        json_encoders = {
            datetime: lambda dt: arrow.get(dt).isoformat(),
            HttpRequest: lambda req: req.dict(),
            HttpResponse: lambda res: res.dict(),
        }


def build_http_request_from_yaml_template(templates_manager=None, template_filename: str = None, ctx=None):
    """
    Build a request from a YAML template.

    :param templates_manager:
    :type:
    :return:
    """
    data = templates_manager.render(template_filename, ctx)
    d = yaml.safe_load(data)
    return HttpRequest(**d)



#!/usr/bin/env python

# https://code.activestate.com/recipes/576586-dot-style-nested-lookups-over-dictionary-based-dat/
from typing import Dict, Type

from deepnox import loggers

LOGGER = loggers.factory(__name__)
""" The module LOGGER. """

class Map(dict):
    """
    Example:
    m = Map({'first_name': 'Eduardo'}, last_name='Pool', age=24, sports=['Soccer'])
    """

    def __init__(self, *args, **kwargs):
        try:
            super(Map, self).__init__(*args, **kwargs)
        except ValueError as e:
            LOGGER.error("ValueError while creating class:`deepnox.utils.maps.Map`", exc_info=e)
            raise CannotConvertToMapTypeError(param=str({"_args": args, "_kwargs": kwargs}))
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v
        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        if type(value) == dict and not isinstance(value, Map):
            self.__setitem__(key, Map(value))
        else:
            super().__setitem__(key, value)
            self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Map, self).__delitem__(key)
        del self.__dict__[key]


class UpperMap(Map):

    def __setitem__(self, key, value):
        if type(value) == dict and not isinstance(value, UpperMap):
            self.__setitem__(key, UpperMap(value))
        else:
            super().__setitem__(key, value)
            self.__dict__.update({key: value})

    def __getattr__(self, attr):
        return self.get(attr.lower())


class CannotConvertToMapTypeError(Exception):
    """
    Exception: cannot convert object to :class:``deepnox.utils.maps.Map`.
    """

    def __init__(self, param, message: str = None) -> None:
        """
        Create a new class:`CannotConvertToMapTypeError` exception.
        :param param: The parameter that cannot be converted.
        :param message: An optional alternative message.
                        Default message is: "The parameter ({param}) must be of type character
                        [dict|`deepnox.utils.maps.Map`] but found: (type({param})={type({param})})."
        """
        message = message or f"The parameter ({param}) must be of type character [dict|`deepnox.utils.maps.Map`] but found: (type({param})={type({param})})."
        self.parameter_name = param
        self.message = message
        super().__init__(self.message)


def to_map(o: Type[Dict] = None):
    """
    Convert a dict or derived class to :class:`deepnox.utils.maps.Map`.

    :param o: The dictionary to convert.
    :type o: dict
    :return: The map based on dictionary. An empty :class:`deepnox.utils.maps.Map` object if parameter is None.
    :rtype: :class:`deepnox.utils.maps.Map`
    :raises :class:`deepnox.utils.maps.CannotConvertToMapTypeError`: The provided paramater is not a dictionary.
    """
    if o is None:
        return Map()
    if isinstance(o, Map):
        return o
    elif isinstance(o, dict):
        return Map(o)
    raise CannotConvertToMapTypeError(param=o)



#!/usr/bin/env python3

"""
Base classes and helpers used by deepnox.tests.services layer.

This file is a part of python-wipbox project.

(c) 2021, Deepnox SAS.
"""


class BaseService(object):
    """
    A deepnox.tests.app class for logic & business deepnox.tests.services.
    """


class BaseAioService(BaseService):
    """ """

    def __init__(self, client: object, bearer_token: str = None):
        self.client = client
        self.bearer_token = bearer_token


class ComputingService(BaseService):
    """
    A deepnox.tests.app class for computing deepnox.tests.services.
    """

    def process(self):
        """
        Process the whished computing.
        """
        raise NotImplementedError("This coro `process` is not implemented.")


class BusinessService(BaseService):
    """
    A deepnox.tests.app class for business deepnox.tests.services.
    """
